import abc
import hashlib
import json
import os
import re
import shutil
import subprocess
import tempfile

try:
    import testcases
except ModuleNotFoundError:
    from util import testcases

DESCFILE    = '../data/descfile.txt'
BINARIES    = '../data/klausur_zweittermin/bin'
OBJECTS     = '../data/klausur_zweittermin/objects'
SUBMISSIONS = '../data/ok.json'
HEADER      = '../data/klausur_zweittermin/code-testing'


def run_cmd(cmd, stdin=None, check=False):
    return subprocess.run(
        'ulimit -v 1024;' + cmd,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        input=stdin,
        shell=True,
        check=check,
        encoding='utf-8',
        timeout=0.5
    )


def all_subclasses(cls):
    return cls.__subclasses__() \
        + [g for s in cls.__subclasses__() for g in all_subclasses(s)]


def sha1(submission_obj):
    return hashlib.sha1(submission_obj['code'].encode()).hexdigest()


class Test(metaclass=abc.ABCMeta):
    """docstring for IliasQuestion"""

    @classmethod
    def available_tests(cls):
        return {sub.__name__ : sub for sub in all_subclasses(cls)}

    def __new__(cls, *args, **kwargs):
        assert hasattr(cls, 'depends'),         "depends not defined"
        assert hasattr(cls, 'label_success'),   "label_success not defined"
        assert hasattr(cls, 'label_failure'),   "label_failure not defined"
        return super().__new__(cls)

    def __init__(self, submission_obj, **kwargs):

        if not self.dependencies_satisfied(submission_obj):
            self.result     = False
            self.annotation = "TEST DEPENDENCY NOT MET"
            self.serialize(submission_obj)

        elif str(self) in submission_obj['tests']:
            self.deserialize(submission_obj['tests'][str(self)])

        else:
            self.result, self.annotation = self.run_test(
                submission_obj, **kwargs)
            self.serialize(submission_obj)

    def __bool__(self):
        return self.result

    def __str__(self):
        return self.__class__.__name__

    def dependencies_satisfied(self, submission_obj):
        return all(dep(submission_obj).result for dep in self.depends)

    def deserialize(self, test):
        self.result     = test['label'] == self.label_success
        self.annotation = test['annotation']

    def serialize(self, submission_obj):
        as_dict = {
            'name'       : str(self),
            'annotation' : self.annotation
        }

        if self.result:
            as_dict['label'] = self.label_success
        else:
            as_dict['label'] = self.label_failure

        submission_obj['tests'][str(self)] = as_dict

    @abc.abstractmethod
    def run_test(self, submission_obj) -> (bool, str):
        return NotImplemented


class EmptyTest(Test):
    """docstring for EmptyTest"""

    depends         = ()
    label_success   = 'NOT_EMPTY'
    label_failure   = 'EMPTY'

    def run_test(self, submission_obj):
        return bool(submission_obj['code'].strip()), ""


class CompileTest(Test):

    depends         = (EmptyTest, )
    label_success   = 'COMPILATION_SUCCESSFUL'
    label_failure   = 'COMPILATION_FAILED'

    def run_test(self, submission_obj):

        ret = run_cmd("gcc-7 -c -xc -Icode-testing -o code.o -",
                      submission_obj['code'])
        return not ret.returncode, ret.stderr


class LinkTest(Test):

    depends         = (CompileTest, )
    label_success   = 'LINKING_SUCCESSFUL'
    label_failure   = 'LINKING_FAILED'

    def run_test(self, submission_obj):

        t = submission_obj['type']
        m = re.search(r'(a0\d)', t)

        ret = run_cmd(f"gcc-7 -o code objects/{m.group(0)}-testing.o code.o")
        return not ret.returncode, ret.stderr


class UnitTestTest(Test):
    """docstring for UnitTestTest"""

    depends         = (LinkTest, )
    label_success   = 'UNITTEST_SUCCSESSFUL'
    label_failure   = 'UNITTEST_FAILED'

    @staticmethod
    def testcase(i, args, stdout):
        try:
            ret = run_cmd("./code %s" % args, check=True)
            assert ret.stdout == stdout
        except AssertionError:
            return False, f"Case #{i}: [ASSERT FAILED] ./program {args} WAS '{ret.stdout}' SHOULD '{stdout}'"
        except subprocess.CalledProcessError as err:
            return False, f"Case #{i}: [FAILED] ./program {args} WITH ERROR '{err.stderr}'"
        except subprocess.TimeoutExpired:
            return False, f"Case #{i}: [TIMEOUT] ./program {args}"
        else:
            return True,  f"Case #{i}: [SUCCESS] ./program {args}"

    def run_test(self, submission_obj):

        task = testcases_dict[submission_obj['type']]
        results, messages = zip(*list(self.testcase(i, case, result)
                                      for i, (case, result) in enumerate(zip(task['cases'], task['results']))))

        return all(results), '\n'.join(messages)


def processing(highest_test):

    with open(SUBMISSIONS) as submission_file:
        submissions = json.JSONDecoder().decode(submission_file.read())

    # Get something disposable
    path = tempfile.mkdtemp()
    run_cmd(f'cp -r {OBJECTS}  {path}')
    run_cmd(f'cp -r {BINARIES} {path}')
    run_cmd(f'cp -r {HEADER} {path}')
    os.chdir(path)

    for username, data in submissions.items():
        for submission_obj in data['submissions']:
            highest_test(submission_obj)
            run_cmd('rm code*')

    shutil.rmtree(path)
    return submissions


if __name__ == '__main__':
    testcases_dict = testcases.evaluated_testcases(DESCFILE)
    print(json.dumps(processing(UnitTestTest), sort_keys=True, indent=4))
